// ex: set tabstop=2 shiftwidth=2 expandtab syn=c:
// $Id: internal_interrupt.apriori_routes.c,v 1.3 2006/10/04 01:30:22 wangf Exp $

/*                                                                      
 * "Copyright (c) 2000-2003 The Regents of the University  of California.  
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs and the author appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Copyright (c) 2002-2003 Intel Corporation
 * All rights reserved.
 *
 * This file is distributed under the terms in the attached INTEL-LICENSE     
 * file. If you do not find these files, copies can be found by writing to
 * Intel Research Berkeley, 2150 Shattuck Avenue, Suite 1300, Berkeley, CA, 
 * 94704.  Attention:  Intel License Inquiry.
 */

/*
 * Authors:  Rodrigo Fonseca
 * Date Last Modified: 2005/05/26
 */


/* Allows events with global knowledge independent of the actual code
   on the motes. This is a mechanism that allows powerful scripting,
   by inserting identified events in the event queue.*/


#include <S4.h>
#include <S4Command.h>
#include "apriori-routes.h"
//#include "apriori-routes-n1000-1000.h"

/* Description: */
/* This simulation routes according to the predefined routes in
 * apriori-routes.h. It expects this .h file to define 
	MSGS_INITIAL_WAIT (seconds)
	MSGS_INTERVAL (seconds)
	N_ROUTES
	uint16_t routes[N_ROUTES][2]
  It does not kill motes.
*/

enum {
  INT_EVENT_FIRST = 0x8000,
  INT_EVENT_SEND_MSG,
};

/* Internal Functions */

//Forward Declarations
#ifdef S4ROUTING
result_t select_determined_routing_pair(short from, short to,
                          Coordinates ** coords_from, Coordinates ** coords_to);
result_t select_random_routing_pair(short *from, short *to, 
     Coordinates ** coords_from, Coordinates ** coords_to);
#endif
#ifdef CRROUTING
result_t select_determined_routing_pair(short from, short to, uint8_t *beacon);
result_t select_random_routing_pair(short *from, short *to, uint8_t *beacon);
#endif
long int get_next_route_interval();
#ifdef S4ROUTING
void send_route_to_command(short origin, short dest, Coordinates * coords_to);
#endif
#ifdef CRROUTING
void send_route_to_command(short origin, short dst, uint8_t beacon);
#endif



void schedule_first_events() {
  dbg_clear(DBG_ROUTE,"II:schedule_first_events, at %llu\n",tos_state.tos_time);
  scheduleInterrupt(INT_EVENT_SEND_MSG, (long long)MSGS_INITIAL_WAIT*4000000);
}

void send_route_to_random_pair() {
  short from,to;
  static int i = 0;
#ifdef S4ROUTING
  Coordinates *coords_from,*coords_to;
#endif
#ifdef CRROUTING
  uint8_t beacon;
#endif
  dbg_clear(DBG_USR3,"II:Choosing next pair to route\n");
  //if (select_random_routing_pair(&from,&to,&coords_from,&coords_to) == SUCCESS) {
  from = routes[i][0]; to = routes[i][1];
#ifdef S4ROUTING
  if (select_determined_routing_pair(from, to,&coords_from,&coords_to) == SUCCESS) {
#endif
#ifdef CRROUTING
  if (select_determined_routing_pair(from,to,&beacon) == SUCCESS) {
#endif
    dbg_clear(DBG_ROUTE,"II:Selected pair: from %d to %d (closest_beacon %d)\n",from,to,beacon);
#ifdef S4ROUTING
    dbg_clear(DBG_USR3,"II:Coordinates for dest mote %d: ",to);
    coordinates_print(DBG_USR3,coords_to);
    //Send route command
    send_route_to_command(from,to,coords_to);
#endif
#ifdef CRROUTING
    send_route_to_command(from,to,beacon);
#endif
  } else {
    dbg_clear(DBG_ROUTE,"II:Failed to select pair, skipping slot\n");
  }
  //if (++i < N_ROUTES)
  //  scheduleInterrupt(INT_EVENT_SEND_MSG, tos_state.tos_time + get_next_route_interval());
  i++;
}

/* Every one second...*/
long int get_next_route_interval() {
  return (long int)4000000*MSGS_INTERVAL;
}

#ifdef S4ROUTING
result_t select_determined_routing_pair(short from, short to,
                          Coordinates ** coords_from, Coordinates ** coords_to) {
#endif
#ifdef CRROUTING
result_t select_determined_routing_pair(short from, short to, uint8_t *beacon) {
#endif
  bool found;
  char *name = "S4StateM$my_coords";
  uintptr_t p;
  size_t size;
#ifdef CRROUTING
  Coordinates * coords_to;
  Coordinates * coords_from;
#endif

  //check that mote is on and has valid coordinates of at least mask

  //destination
  if (__nesc_nido_resolve(to, name, &p, &size) != 0) {
    dbg(DBG_ROUTE,"II: Error resolving %s for mote %d\n",name,to);
    return FAIL;
  }
#ifdef S4ROUTING
  *coords_to = (Coordinates*)p;
  found = (tos_state.moteOn[to] && 
           coordinates_count_valid_components(*coords_to) == N_ROOT_BEACONS);
#endif
#ifdef CRROUTING
  coords_to = (Coordinates*)p;
  found = tos_state.moteOn[to];
  //found = (tos_state.moteOn[to] &&
  //         coordinates_count_valid_components(coords_to) == N_ROOT_BEACONS);
  if (found) {
    *beacon = coordinates_get_closest_beacon(coords_to);
    //coordinates_print(DBG_ROUTE,coords_to);
  }
#endif
  if (!found) {
#ifdef S4ROUTING
    dbg_clear(DBG_ROUTE,"II: Invalid destination %d: off or invalid coordinates\n", to);
    dbg_clear(DBG_ROUTE,"II: number of valid components %d\n", coordinates_count_valid_components(coords_to));
#endif
#ifdef CRROUTING
    dbg_clear(DBG_ROUTE,"II: mote %d is not on!\n", to);
#endif
    return FAIL;
  }
  //source
  if (__nesc_nido_resolve(from, name, &p, &size) != 0) {
    dbg(DBG_ROUTE,"II: Error resolving %s for mote %d\n",name,from);
    return FAIL;
  }
#ifdef S4ROUTING
  *coords_from = (Coordinates*)p;
  found = (tos_state.moteOn[from] && 
           coordinates_count_valid_components(*coords_from) == N_ROOT_BEACONS);
#endif
#ifdef CRROUTING
  coords_from = (Coordinates*)p;
  found = tos_state.moteOn[from];
  //found = (tos_state.moteOn[from] &&
  //         coordinates_count_valid_components(coords_from) == N_ROOT_BEACONS);
#endif
 
  if (!found) {
#ifdef S4ROUTING
    dbg_clear(DBG_ROUTE,"II: Invalid source %d: off or invalid coordinates\n", from);
    dbg_clear(DBG_ROUTE,"II: number of valid components %d\n", coordinates_count_valid_components(coords_from));
#endif
#ifdef CRROUTING
    dbg_clear(DBG_ROUTE,"II: mote %d is not on!\n", from);
#endif
    return FAIL;
  }
  return SUCCESS;
}

#ifdef S4ROUTING
result_t select_random_routing_pair(short *from, short *to,
                          Coordinates ** coords_from, Coordinates ** coords_to) {
#endif
#ifdef CRROUTING
result_t select_random_routing_pair(short *from, short *to, uint8_t *beacon) {
#endif
  //select random dest
  int i,c,d;
  bool found;
  char *name = "S4StateM$my_coords";
  uintptr_t p;
  size_t size;
#ifdef CRROUTING
  Coordinates * coords_to;
  Coordinates * coords_from;
#endif
  
  d = (int)(((double)tos_state.num_nodes)*rand()/(RAND_MAX+1.0));
  //check that mote is on and has valid coordinates of at least mask
  found = 0;
  for (i = 0; i < tos_state.num_nodes && !found; i++) {
    c = (d + i) % tos_state.num_nodes;
    //get coordinates for this mote and count valid components 
    if (__nesc_nido_resolve(c, name, &p, &size) != 0) {
      dbg(DBG_USR3,"II: Error resolving %s for mote %d\n",name,c);
      continue;
    }
#ifdef S4ROUTING
    *coords_to = (Coordinates*)p;
    found = (tos_state.moteOn[c] && 
             coordinates_count_valid_components(*coords_to) == N_ROOT_BEACONS);
    if (found) 
      *to = c;
#endif
#ifdef CRROUTING
    coords_to = (Coordinates*)p;
    found = (tos_state.moteOn[c] &&
             coordinates_count_valid_components(coords_to) == N_ROOT_BEACONS);
    if (found) {
        *to = c;
        *beacon = coordinates_get_closest_beacon(coords_to);
      }
#endif
  }
  if (!found)
    return FAIL;
  //look for a source
  d = (int)(((double)tos_state.num_nodes)*rand()/(RAND_MAX+1.0));
  //check that mote is on and has valid coordinates of at least mask

  found = 0;
  for (i = 0; i < tos_state.num_nodes && !found; i++) {
    c = (d + i) % tos_state.num_nodes;
    if (c == *to) continue;
    //get coordinates for this mote and count valid components 
    if (__nesc_nido_resolve(c, name, &p, &size) != 0) {
      dbg(DBG_USR3,"II: Error resolving %s for mote %d\n",name,c);
      continue;
    }
#ifdef S4ROUTING
    *coords_from = (Coordinates*)p;
    found = (tos_state.moteOn[c] && 
             coordinates_count_valid_components(*coords_from) == N_ROOT_BEACONS);
#endif
#ifdef CRROUTING
    coords_from = (Coordinates*)p;
    found = (tos_state.moteOn[c] &&
             coordinates_count_valid_components(coords_from) == N_ROOT_BEACONS);
#endif

    if (found) 
      *from = c;
  }
  if (!found)
    return FAIL;
  return SUCCESS;
}

#ifdef S4ROUTING
void send_route_to_command(short from, short to,Coordinates *coords) {
#endif
#ifdef CRROUTING
void send_route_to_command(short from, short to, uint8_t beacon) {
#endif
  TOS_Msg message;
  S4CommandMsg *cm = (S4CommandMsg*) &message.data;
  
  S4CommandArgs* cmd_args = 
       (S4CommandArgs*) &cm->type_data.data;
  
  //added by Feng Wang on Sept. 29, to increment cmd seqno
  static int i = 0;
  
  message.addr = from;
  message.type = AM_S4_COMMAND_MSG;
  message.group = TOS_AM_GROUP; 
  message.length = sizeof(S4CommandMsg);
  
  cm->header.last_hop = TOS_UART_ADDR;
  cm->header.seqno = 0;
  
  cm->type_data.hopcount = 1;
  cm->type_data.origin = TOS_UART_ADDR;
  cm->type_data.type = S4_CMD_APP_ROUTE_TO;

  cmd_args->seqno = ++i;
  //avoid 0 since "last_seen_seqno" is initialized to 0
  if (cmd_args->seqno == 0)
     cmd_args->seqno = 1;
  cmd_args->flags = 0;

  //Destination
  cmd_args->args.dest.addr = to;
  //cmd_args->args.dest.mode = 2;
#ifdef S4ROUTING
  coordinates_copy(coords, &cmd_args->args.dest.coords);
#endif
#ifdef CRROUTING
  cmd_args->args.dest.closest_beacon = beacon;
#endif

  //Done setting up message, call 
  sendUARTMessage(from, &message);
  
}


/* --------------------------------------------------------------------*/
/* Exported Functions */

void internalInterruptInit() {
  //dbg_clear(DBG_ROUTE,"II:InternalInterruptInit! %llu\n",tos_state.tos_time);
  scheduleInterrupt(INT_EVENT_FIRST, tos_state.tos_time);
}


/*This function is called whenever an interrupt event is
 *triggered. This allows us to dispatch different handlers
 *based on id. Rescheduling can be done by calling
 *scheduleInterrupt(id,time)
 */
void internalInterruptHandler(uint32_t id) {
#ifdef MULTIPLE_FLOWS
  uint8_t flow;
#endif
  static int total_flows = 0;
  //dbg_clear(DBG_ROUTE,"II:InternalInterruptHandler %llu!\n",tos_state.tos_time);
  switch(id) {
    case INT_EVENT_FIRST:
      schedule_first_events();
      break;
    case INT_EVENT_SEND_MSG:
#ifndef MULTIPLE_FLOWS
      send_route_to_random_pair();
      total_flows++;
#else
      for (flow=0; flow<MULTIPLE_FLOWS; flow++)
        send_route_to_random_pair();
      total_flows += MULTIPLE_FLOWS;
#endif
      if (total_flows < N_ROUTES)
        scheduleInterrupt(INT_EVENT_SEND_MSG, tos_state.tos_time + get_next_route_interval());
      break;
    default:
      break;
  }
}


