import commands, time

DEFAULT_RETRY=5

COUNT_NUM = 10
RSSI_NUM = 10

def run_retry(cmd, retry=DEFAULT_RETRY, print_cmd = True, default_output = ''):
    """run the command until the exit code is 0, or exceed the maximum retry setting
    """
    real_output = default_output
    if print_cmd:
        print cmd
    for i in range(retry):
        (status, output) = commands.getstatusoutput(cmd)
        if print_cmd:
            print "(%d) %s"% (status, output)
        if status == 0:
            real_output = output
            break

    return (status, real_output)


class GatewayMap:
    mapping = {}
    gateway_num = 0

    def __init__(self, gw_file):
        for line in open(gw_file):
            if not line: continue
            if line[0] == '#': continue
            args = line.split()
            arg_type = args[0]
            if arg_type=="gateway_num":
                self.gateway_num = int(args[1])
            elif arg_type=="mote":
                gw_id = int(args[1])
                mote_id = int(args[2])
                if gw_id >0:
                    self.mapping[mote_id] = gw_id
                else:
                    print 'ignore mote', mote_id,'with 0 gateway_id'
                
measure_template = "java net.tinyos.cr.TestMeasure  %d %d measure %d %d %d %d"
count_template = "java net.tinyos.cr.TestMeasure %d %d count %d"
clear_template = "java net.tinyos.cr.TestMeasure %d %d clear"
version_template = "java net.tinyos.cr.TestMeasure %d %d version"
rssi_template = "java net.tinyos.cr.TestMeasure %d %d rssi %d"

class MeasureTestbed(GatewayMap):
    measure_seq  = []
    def __init__(self, gw_file):
        GatewayMap.__init__(self, gw_file)
        max_mote_id = max(self.mapping.keys())
        self.measure_seq = [0]*(max_mote_id+1)
        
    def measure(self, source, n_packets, interval, power_level, retry=5):
        self.measure_seq[source] += 1
        cmd = measure_template %(self.mapping[source], source, n_packets, interval, power_level, self.measure_seq[source])
        res = run_retry(cmd, retry=retry)
        return res

    def count(self, mote_id, broadcast_source, retry=5):
        cmd = count_template % (self.mapping[mote_id], mote_id, broadcast_source)
        res = run_retry(cmd, retry=retry)
        return res

    def rssi(self, mote_id, broadcast_source, retry=5):
        cmd = rssi_template % (self.mapping[mote_id], mote_id, broadcast_source)
        res = run_retry(cmd, retry=retry)
        return res

    def clear(self, mote_id, retry=5):
        cmd = clear_template % (self.mapping[mote_id], mote_id)
        res = run_retry(cmd, retry=retry)
        if res[0]<>0:
            self.measure_seq[mote_id] = 0
        return res

    def version(self, mote_id, retry=5):
        cmd = version_template % (self.mapping[mote_id], mote_id)
        return run_retry(cmd, retry=retry)

    def clear_all(self, retry=5):
        res = []
        for mote_id in self.mapping.keys():
            (status, tmp) = self.clear(mote_id, retry)
            res.append(status)
        return res

    def version_all(self, retry=5):
        res = []
        for mote_id in self.mapping.keys():
            (status, output) = self.version(mote_id, retry)
            res.append(output)
        return res
            
    def measure_all(self, n_packets, interval, power_level, sleep_interval, retry=5):
        packet_sent = []
        for src in self.mapping.keys():
            (status, output) =self.measure(src, n_packets, interval, power_level, retry)
            if status == 0:
                packet_sent.append(n_packets)
            else:
                packet_sent.append(0)
            print 'sleeping', sleep_interval+n_packets*interval/1000, 'seconds'
            time.sleep(sleep_interval +n_packets*interval/1000)

        return packet_sent
    
    def count_all(self, matrix_fn, retry=5):
        fd = open(matrix_fn, 'w')

        max_mote_id = max(self.mapping.keys())
        
        #for mote_id in self.mapping.keys():
        #    fd.write('%d ' %(mote_id,))
        #fd.write('\n')

            
        for mote_id in self.mapping.keys():
            counts_str = []
            for broadcast_src in range(1, max_mote_id+1, COUNT_NUM):
                (status, output) = self.count(mote_id, broadcast_src, retry)
                if status==0:
                    counts_str.extend( output.strip().split() )
                else:
                    counts_str.extend( ['-1']*COUNT_NUM)
                    
            for src in self.mapping.keys():
                fd.write('%s ' %( counts_str[src-1], ))
            fd.write('\n')
            
        fd.close()
    
    def rssi_all(self, matrix_fn, retry=5):
        fd = open(matrix_fn, 'w')

        max_mote_id = max(self.mapping.keys())
        
        #for mote_id in self.mapping.keys():
        #    fd.write('%d ' %(mote_id,))
        #fd.write('\n')

            
        for mote_id in self.mapping.keys():
            rssis_str = []
            for broadcast_src in range(1, max_mote_id+1, RSSI_NUM):
                (status, output) = self.rssi(mote_id, broadcast_src, retry)
                if status==0:
                    rssis_str.extend( output.strip().split() )
                else:
                    rssis_str.extend( ['-1']*RSSI_NUM)
                    
            for src in self.mapping.keys():
                fd.write('%s ' %( rssis_str[src-1], ))
            fd.write('\n')
            
        fd.close()
        
