// ex: set tabstop=2 shiftwidth=2 expandtab syn=c:
// $Id: coordinates.h,v 1.4 2006/03/14 22:57:58 wangf Exp $
                                    
/*                                                                      
 * "Copyright (c) 2000-2003 The Regents of the University  of California.  
 * All rights reserved.             
 *                                  
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs and the author appear in all copies of this software.
 *                                  
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *                                  
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *                                  
 * Copyright (c) 2002-2003 Intel Corporation
 * All rights reserved.             
 *                                  
 * This file is distributed under the terms in the attached INTEL-LICENSE     
 * file. If you do not find these files, copies can be found by writing to
 * Intel Research Berkeley, 2150 Shattuck Avenue, Suite 1300, Berkeley, CA, 
 * 94704.  Attention:  Intel License Inquiry.
 */                                 
                                    
/*                                  
 * Authors:  Rodrigo Fonseca        
 * Date Last Modified: 2005/05/26
 */

/*
 * Authors:  Feng Wang, Univ. of Texas at Austin, CS Department
 * Date Last Modified: 2006/02/04
 */

/* This file defines coordinates to be a fixed sized array of
 * 8 bit numbers. 255 indicates that a given component is not
 * valid. 
 * IMPORTANT: Beware that when changing the number of components, it may be
 * necessary to change the size of TOS_Msg*/

#ifndef COORDS_H
#define COORDS_H

#include "util.h"


/* interface:
These are "exported" from this file. The implementation is subject to change.
inline result_t coordinates_init(Coordinates* c)
inline result_t coordinates_copy(Coordinates* source, Coordinates* dest)
inline uint8_t coordinates_copy_k_closest(Coordinates* source, Coordinates* dest, uint8_t K)
inline result_t coordinates_set_component(Coordinates* c, uint8_t pos, uint8_t hopcount)
inline uint16_t coordinates_distance(Coordinates* c1, Coordinates* c2) 
inline uint8_t coordinates_get_closest_beacon(Coordinates* c) {
inline void coordinates_print(int debug,Coordinates* c) 
inline bool coordinates_is_valid_component(Coordinates* c, uint8_t pos)
inline bool coordinates_has_same_valid_components(Coordinates *q, Coordinates *c) {
inline uint8_t coordinates_count_valid_components(Coordinates* c)
*/

#ifndef INVALID_NODE_ID
#ifdef PLATFORM_PC
#define INVALID_NODE_ID 65535
#else
#define INVALID_NODE_ID 255
#endif
#endif

/*
#ifndef MAX_ROOT_BEACONS
#ifdef PLATFORM_PC
#define MAX_ROOT_BEACONS 30
#else
#define MAX_ROOT_BEACONS 8
#endif
#endif
*/

enum {
  INVALID_DISTANCE = 65535u,
  MAX_COORD_DISTANCE = 16384u,     
  MAX_COMPONENT_VALUE = 254,
  INVALID_COMPONENT = 255,
};

typedef struct Coordinates {
  uint8_t comps[MAX_ROOT_BEACONS];
} __attribute__ ((packed)) Coordinates, *Coordinates_ptr;


/* declaration */
#ifdef PLATFORM_PC
inline void coordinates_print(int debug,Coordinates* c);
#else
#define coordinates_print(a,b) 
#endif


inline result_t coordinates_init(Coordinates* c) {
  int i;
  if (c == NULL) 
    return FAIL;
  for (i = 0; i < MAX_ROOT_BEACONS; c->comps[i] = INVALID_COMPONENT, i++);
  return SUCCESS;
}

inline bool coordinates_is_valid_component(Coordinates* c, uint8_t pos) {
  if (c == NULL || pos >= MAX_ROOT_BEACONS)
    return FALSE;
  return !(c->comps[pos] == INVALID_COMPONENT);
}



inline uint8_t coordinates_count_valid_components(Coordinates* c) {
  uint8_t i;
  uint8_t count = 0;
  if (c == NULL) 
    return 0;
  for (i = 0; i < MAX_ROOT_BEACONS; i++) {
    if (coordinates_is_valid_component(c, i))
      count++;
  }
  return count;
}

inline result_t coordinates_copy(Coordinates* source, Coordinates* dest) {
  int i;
  if (source == NULL || dest == NULL) 
    return FAIL;
  for (i = 0; i < MAX_ROOT_BEACONS; i++)
    dest->comps[i] = source->comps[i];
  return SUCCESS;
}

inline result_t coordinates_set_component(Coordinates* c, uint8_t pos, uint8_t hopcount) {
  dbg(DBG_USR2,"coordinates_set_component\n");
  if (c == NULL || pos >= MAX_ROOT_BEACONS) 
    return FAIL;
  c->comps[pos] = hopcount;
  return SUCCESS;
}

/* Given coords, returns the beacon that is closest to it, i.e., the
   index of the smallest component */
inline uint8_t coordinates_get_closest_beacon(Coordinates* c) {
  uint8_t i;
  uint8_t closest = INVALID_COMPONENT;
  uint8_t closest_distance = INVALID_COMPONENT;
  if (c == NULL) 
    return INVALID_COMPONENT;
  for (i = 0; i < MAX_ROOT_BEACONS; i++) {
    if (coordinates_is_valid_component(c, i)) {
      if (c->comps[i] < closest_distance) {
        closest = i;
        closest_distance = c->comps[i];
      }
    }
  }
  return closest; 
}

#ifdef PLATFORM_PC
inline void coordinates_print(int debug,Coordinates* c) {
  int i;
  dbg_clear(debug,"%d",coordinates_is_valid_component(c,0)?(c->comps[0]):-1);
  //for (i = 1; i < MAX_ROOT_BEACONS; i++) {
  for (i = 1; i < N_ROOT_BEACONS; i++) {
    dbg_clear(debug,", %d",coordinates_is_valid_component(c,i)?(c->comps[i]):-1);
  }
  dbg_clear(debug,"\n");
}
#endif
#endif
